import bs
import bsTeamGame
import bsMap
import bsUtils
import bsSpaz
import bsScoreBoard
import weakref
import bsAnnounce
import bsVector

def getGames():
    return [EliminationGame]


class Icon(bs.Actor):
    def __init__(self,activity,retain,player,position,scale,showLives=True,showDeath=True,nameScale=1.0):
        bs.Actor.__init__(self,activity,retain)

        self._player = weakref.ref(player)
        self._showLives = showLives
        self._showDeath = showDeath
        self._nameScale = nameScale
        icon = player.getIcon()
        self.node = bs.newNode('image',
                               owner=self,
                               attrs={'texture':icon['texture'],
                                      'tintTexture':icon['tintTexture'],
                                      'tintColor':icon['tintColor'],
                                      'tint2Color':icon['tint2Color'],
                                      'opacity':1.0,
                                      'absoluteScale':True,
                                      'attach':'bottomCenter'})

        self._nameText = bs.newNode('text',
                                    owner=self.node,
                                    attrs={'text':player.getName(),
                                           'color':bsUtils.getSafeDisplayColor(player.getTeam().color),
                                           'hAlign':'center',
                                           'vAlign':'center',
                                           'maxWidth':120.0,
                                           'hAttach':'center',
                                           'vAttach':'bottom'})

        if self._showLives:
            self._livesText = bs.newNode('text',
                                         owner=self.node,
                                         attrs={'text':'x0',
                                                'color':(1,1,0.5),
                                                'hAlign':'left',
                                                'hAttach':'center',
                                                'vAttach':'bottom'})
        self.setPositionAndScale(position,scale)

    def setPositionAndScale(self,position,scale):
        self.node.translate = position
        self.node.scale = [70.0*scale]
        self._nameText.translate = (position[0],position[1]+scale*35.0)
        self._nameText.scale = [0.8*scale*self._nameScale]
        if self._showLives:
            self._livesText.translate = (position[0]+scale*10.0,position[1]-scale*43.0)
            self._livesText.scale = [1.0*scale]

    def updateForLives(self):
        lives = self._player().data['lives']
        if self._showLives:
            if lives > 0: self._livesText.text = 'x'+str(lives-1)
            else: self._livesText.text = ''
        if lives == 0:
            self._nameText.opacity = 0.2
            self.node.color = (0.7,0.3,0.3)
            self.node.opacity = 0.2
        
    def handlePlayerSpawned(self):
        if not self.node.exists(): return
        self.node.opacity = 1.0
        self.updateForLives()

    def handlePlayerDied(self):
        if not self.node.exists(): return
        if self._showDeath:
            bsUtils.animate(self.node,'opacity',{0:1.0,50:0.0,100:1.0,150:0.0,200:1.0,250:0.0,
                                                 300:1.0,350:0.0,400:1.0,450:0.0,500:1.0,550:0.2})
            bs.gameTimer(600,bs.WeakCall(self.updateForLives))
        

class EliminationGame(bsTeamGame.TeamGameActivity):

    @classmethod
    def getName(c):
        return 'Elimination'

    @classmethod
    def getDescription(c,sessionType):
        return 'Last remaining alive wins.'

    @classmethod
    def getSupportedSessionTypes(self):
        return ['freeForAll','teams']

    @classmethod
    def getSupportedMaps(self,sessionType):
        return bsMap.getMapList("melee")

    @classmethod
    def getSettings(self,sessionType):
        settings = [("Lives Per Player",{'default':1,'minValue':1,'maxValue':10,'increment':1}),
                    ("Time Limit",{'choices':[('None',0),('1 Minute',60),
                                            ('2 Minutes',120),('5 Minutes',300),
                                            ('10 Minutes',600),('20 Minutes',1200)],'default':0}),
                    ("Respawn Times",{'choices':[('Shorter',0.25),('Short',0.5),('Normal',1.0),('Long',2.0),('Longer',4.0)],'default':1.0}),
                    ("Epic Mode",{'default':False})]
        if sessionType == 'teams':
            settings.append(("Solo Mode",{'default':False}))
            settings.append(("Balance Total Lives",{'default':False}))
        return settings

    @classmethod
    def getConfigDescription(self,config):
        name = self.getNameTranslated()
        if 'Solo Mode' in config['settings'] and config['settings']['Solo Mode']:
            name = bs.getResource('soloNameFilterText').replace('${NAME}',name)
        if 'Epic Mode' in config['settings'] and config['settings']['Epic Mode']:
            name = bs.getResource('epicNameFilterText').replace('${NAME}',name)
        s = name+" @ "+bsMap.getTranslatedMapName(bsMap.getFilteredMapName(config['map']))
        return s

    def __init__(self,session,level,settings):
        bsTeamGame.TeamGameActivity.__init__(self,session,level,settings)
        if self.settings['Epic Mode']: self._isSlowMotion = True
        try: self._soloMode = settings['Solo Mode']
        except Exception: self._soloMode = False
        self._scoreBoard = bsScoreBoard.ScoreBoard(self)
        self._staticPlayerData = {}

    def getInstanceName(self):
        name = self.getNameTranslated()
        if self._soloMode: name = bs.getResource('soloNameFilterText').replace('${NAME}',name)
        return self.standardInstanceNameFilter(name)

    def getInstanceDescription(self):
        return 'Last team standing wins.' if self.getSession().getSessionType() == 'teams' else 'Last one standing wins.'

    def getInstanceScoreBoardDescription(self):
        return 'last team standing wins' if self.getSession().getSessionType() == 'teams' else 'last one standing wins'

    def onTransitionIn(self):
        bsTeamGame.TeamGameActivity.onTransitionIn(self, music='Epic' if self.settings['Epic Mode'] else 'Survival')
        self._startGameTime = bs.getGameTime()

    def onTeamJoined(self,team):
        team.data['survivalSeconds'] = None
        team.data['spawnOrder'] = []

    def onPlayerJoined(self,player):
        # we recycle player data so that joining/leaving won't buy you a new set of lives
        if player.playerID in self._staticPlayerData: player.data = self._staticPlayerData[player.playerID]
        else:
            # first time player
            player.data['lives'] = self.settings['Lives Per Player']
        
        if self._soloMode:
            player.data['icons'] = []
            player.getTeam().data['spawnOrder'].append(weakref.ref(player))
            self._updateSoloMode()
        else:
            # create our icon and spawn
            player.data['icons'] = [Icon(self,bs.Actor.UNRETAINED,player,position=(0,50),scale=0.8)]
            if player.data['lives'] > 0:
                self.spawnPlayer(player)

        # dont waste time doing this until begin
        if self.hasBegun():
            self._updateIcons()

    def _updateSoloMode(self):
        # for both teams, find the first player on the spawn order list with lives remaining
        # and spawn them if they're not alive
        for team in self.teams.values():
            # prune dead players from the spawn order
            team.data['spawnOrder'] = [p for p in team.data['spawnOrder'] if p() is not None]
            for player in [pRef() for pRef in team.data['spawnOrder']]:
                if player.data['lives'] > 0:
                    if not player.isAlive(): self.spawnPlayer(player)
                    break

    def _updateIcons(self):
        # in free-for-all mode, everyone is just lined up along the bottom
        if self.getSession().getSessionType() == 'freeForAll':
            count = len(self.teams)
            xOffs = 70
            x = xOffs*(count-1) * -0.5
            for i,team in enumerate(self.teams.values()):
                if len(team.players) > 1:
                    print 'WTF have',len(team.players),'players in ffa team'
                elif len(team.players) == 1:
                    player = team.players.values()[0]
                    if len(player.data['icons']) != 1:
                        print 'WTF have',len(player.data['icons']),'icons in non-solo elim'
                    for icon in player.data['icons']:
                        icon.setPositionAndScale((x,30),0.7)
                        icon.updateForLives()
                    x += xOffs

        # in teams mode we split up teams
        else:
            if self._soloMode:
                # first off, clear out all icons
                for player in self.players.values():
                    player.data['icons'] = []
                # now for each team, cycle through our available players adding icons
                for team in self.teams.values():
                    if team.teamID == 0:
                        x = -60
                        xOffs = -74
                    else:
                        x = 60
                        xOffs = 74
                    isFirst = True
                    testLives = 1
                    while True:
                        playersWithLives = [p() for p in team.data['spawnOrder'] if p() is not None and p().data['lives'] >= testLives]
                        if len(playersWithLives) == 0: break
                        for player in playersWithLives:
                            player.data['icons'].append(Icon(self,bs.Actor.UNRETAINED,
                                                             player,
                                                             position=(x,(40 if isFirst else 25)),
                                                             scale=1.0 if isFirst else 0.5,
                                                             nameScale=1.0 if isFirst else 0.7,
                                                             showDeath=True if isFirst else False,
                                                             showLives=False))
                            x += xOffs * (0.8 if isFirst else 0.56)
                            isFirst = False
                        testLives += 1
            # non-solo mode
            else:
                for team in self.teams.values():
                    if team.teamID == 0:
                        x = -50
                        xOffs = -70
                    else:
                        x = 50
                        xOffs = 70
                    for player in team.players.values():
                        if len(player.data['icons']) != 1:
                            print 'WTF have',len(player.data['icons']),'icons in non-solo elim'
                        for icon in player.data['icons']:
                            icon.setPositionAndScale((x,30),0.7)
                            icon.updateForLives()
                        x += xOffs
                    
    def _getSpawnPoint(self,player):
        # in solo-mode, if there's an existing live player on the map, spawn at whichever
        # spot is farthest from them (keeps the action spread out)
        if self._soloMode:
            livingPlayer = None
            for team in self.teams.values():
                for player in team.players.values():
                    if player.isAlive():
                        p = player.actor.node.translate
                        # ew.. a freshly spawned player will have a translate of [0,0,0] since they havn't been moved
                        # into place yet - lets ignore those
                        if all(abs(foo) < 0.0001 for foo in p):
                            continue
                        livingPlayer = player
                        livingPlayerPos = p
                        break
            if livingPlayer:
                playerPos = bsVector.Vector(*livingPlayerPos)
                points = []
                for team in self.teams.values():
                    startPos = bsVector.Vector(*self._level.getStartPosition(team.teamID))
                    points.append([(startPos-playerPos).length(),startPos])
                points.sort()
                return points[-1][1]
            else:
                return None
        else:
            return None

        
    def spawnPlayer(self,player):
        
        self.spawnSpazForPlayer(player,self._getSpawnPoint(player))
        if not self._soloMode:
            bs.gameTimer(300,bs.WeakCall(self._printLives,weakref.ref(player)))

        # if we have any icons, update their state
        for icon in player.data['icons']:
            icon.handlePlayerSpawned()

    def _printLives(self,playerRef):
        try:
            player = playerRef()
            pos = playerRef().actor.node.translate
        except Exception: return
        bsAnnounce.AchievementText(self,bs.Actor.RETAINED,
                                   'x'+str(player.data['lives']-1),color=(1,1,0,1),
                                   offset=(0,-0.8,0),jitter=0.0,scale=1.8,position=pos)

    def onPlayerLeft(self,player):
        bsTeamGame.TeamGameActivity.onPlayerLeft(self,player)

        # store this player's data so we can recycle it if he rejoins
        self._staticPlayerData[player.playerID] = player.data

        # still wanna kill our icons though..
        player.data['icons'] = None

        # remove us from spawn-order
        if self._soloMode:
            player.getTeam().data['spawnOrder'].remove(weakref.ref(player))

        # update icons in a moment since our team will be gone from the list then
        bs.gameTimer(0,bs.WeakCall(self._updateIcons))

    def onBegin(self):
        bsTeamGame.TeamGameActivity.onBegin(self,nameSound=bs.getSound('announceElimination'))
        self.setupStandardTimeLimit(self.settings['Time Limit'])

        if self._soloMode:
            self._vsText = bs.NodeActor(self,bs.Actor.UNRETAINED,
                                        "text",{'translate':(0,81),'hAttach':"center",'hAlign':'center',
                                                'scale':[0.5],'vAttach':"bottom",'color':(0.8,0.8,0.3,1.0),'text':bs.getResource('vsText')})

        # if balance-team-lives is on, add lives to the smaller team until total lives match
        if (self.getSession().getSessionType() == 'teams'
            and self.settings['Balance Total Lives']
            and len(self.teams[0].players) > 0
            and len(self.teams[1].players) > 0):
            if self._getTotalTeamLives(self.teams[0]) < self._getTotalTeamLives(self.teams[1]):
                lesserTeam = self.teams[0]
                greaterTeam = self.teams[1]
            else:
                lesserTeam = self.teams[1]
                greaterTeam = self.teams[0]
            addIndex = 0
            while self._getTotalTeamLives(lesserTeam) < self._getTotalTeamLives(greaterTeam):
                lesserTeam.players.values()[addIndex].data['lives'] += 1
                addIndex = (addIndex + 1) % len(lesserTeam.players)

        self._updateIcons()

    def _getTotalTeamLives(self,team):
        return sum(player.data['lives'] for player in team.players.values())

    def handleMessage(self,m):
        if isinstance(m,bsSpaz.PlayerSpazDeathMessage):
            bsTeamGame.TeamGameActivity.handleMessage(self,m) # augment standard behavior
            player = m.spaz.getPlayer()
            if player is None:
                return # might have left or something.. should we check for end-game here?

            player.data['lives'] -= 1
            if player.data['lives'] < 0:
                print 'Got lives < 0 in Elim; this shouldnt happen'
                player.data['lives'] = 0

            # if we have any icons, update their state
            for icon in player.data['icons']:
                icon.handlePlayerDied()

            # play big death sound on our last death or for every one in solo mode
            if self._soloMode or player.data['lives'] == 0:
                bs.playSound(bsSpaz.getFactory().singlePlayerDeathSound)

            # if we hit zero lives we're dead and the game might be over
            if player.data['lives'] == 0:

                # if the whole team is dead, make note of how long they lasted
                if all(teammate.data['lives'] == 0 for teammate in player.getTeam().players.values()):

                    # log the team survival if we're the last player on the team
                    player.getTeam().data['survivalSeconds'] = (bs.getGameTime()-self._startGameTime)/1000

                    # if someone has won, set a timer to end shortly
                    # (allows the dust to settle and draws to occur if deaths are close enough)
                    if len(self._getLivingTeams()) < 2:
                        self._roundEndTimer = bs.Timer(500,bs.WeakCall(self.endGame))
            # we still have lives; yay!
            else:
                # just respawn in regular mode, or spawn the next guy in line in solo
                if not self._soloMode:
                    self.respawnPlayer(player)

            # in solo, put ourself at the back of the spawn order and let the next dude respawn
            if self._soloMode:
                r = weakref.ref(player)
                player.getTeam().data['spawnOrder'].remove(r)
                player.getTeam().data['spawnOrder'].append(r)
                bs.gameTimer(1000,bs.WeakCall(self._updateSoloAndIcons))

    def _updateSoloAndIcons(self):
        self._updateSoloMode()
        self._updateIcons()

    def _getLivingTeams(self):
        return [team for team in self.teams.values() if len(team.players) > 0 and any(player.data['lives'] > 0 for player in team.players.values())]

    def endGame(self):
        results = bsTeamGame.Results(scoreName='Survived',scoreType='seconds',noneIsWinner=True)

        self._vsText = None # kill our 'vs' if its there

        livingTeams = self._getLivingTeams()

        for t in self.teams.values(): results.setTeamScore(t,t.data['survivalSeconds'])
        self.end(results=results)
        
